@@ -41,9 +41,11 @@ class MyRequestHandlerForTests(MyRequestHandler):
         self._headers = {}
         self.log = []
         self.server = Mock()
-        self.server.renderer.rest_to_html = lambda data, mtime=None: \
+        self.server.renderer.command = None
+        self.server.renderer.watch = None
+        self.server.renderer.rest_to_html = lambda data, mtime=None, filename=None: \
             unicode('HTML for %s with AJAX poller for %s' % (data, mtime))
-        self.server.renderer.render_exception = lambda title, error, source: \
+        self.server.renderer.render_exception = lambda title, error, source, mtime=None: \
             unicode('HTML for error %s: %s: %s' % (title, error, source))
     def send_response(self, status):
         self.status = status
@@ -89,7 +91,7 @@ class TestMyRequestHandler(unittest.TestCase):
         handler = MyRequestHandlerForTests()
         handler.path = '/a.txt'
         handler.server.renderer.root = self.filepath('file.txt')
-        handler.handle_rest_file = lambda fn: 'HTML for %s' % fn
+        handler.handle_rest_file = lambda fn, watch=None: 'HTML for %s' % fn
         handler.wfile = StringIO()
         handler.do_GET()
         self.assertEqual(handler.wfile.getvalue(),
@@ -99,7 +101,7 @@ class TestMyRequestHandler(unittest.TestCase):
         handler = MyRequestHandlerForTests()
         handler.path = '/a.txt'
         handler.server.renderer.root = self.filepath('file.txt')
-        handler.handle_rest_file = lambda fn: 'HTML for %s' % fn
+        handler.handle_rest_file = lambda fn, watch=None: 'HTML for %s' % fn
         handler.wfile = StringIO()
         handler.do_HEAD()
         self.assertEqual(handler.wfile.getvalue(), '')
@@ -109,7 +111,7 @@ class TestMyRequestHandler(unittest.TestCase):
         handler.path = '/'
         handler.server.renderer.root = self.filepath('file.txt')
         handler.server.renderer.command = None
-        handler.handle_rest_file = lambda fn: 'HTML for %s' % fn
+        handler.handle_rest_file = lambda fn, watch=None: 'HTML for %s' % fn
         with patch('os.path.isdir', lambda dir: False):
             body = handler.do_GET_or_HEAD()
         self.assertEqual(body, 'HTML for %s' % self.filepath('file.txt'))
@@ -138,7 +140,7 @@ class TestMyRequestHandler(unittest.TestCase):
         handler.path = '/'
         handler.server.renderer.root = None
         handler.server.renderer.command = 'cat README.rst'
-        handler.handle_command = lambda cmd: 'Output of %s' % cmd
+        handler.handle_command = lambda cmd, watch: 'Output of %s' % cmd
         body = handler.do_GET_or_HEAD()
         self.assertEqual(body, 'Output of cat README.rst')
 
@@ -146,7 +148,7 @@ class TestMyRequestHandler(unittest.TestCase):
         handler = MyRequestHandlerForTests()
         handler.path = '/polling?pathname=a.txt&mtime=12345'
         handler.server.renderer.root = self.root
-        handler.handle_polling = lambda fn, mt: 'Got update for %s since %s' % (fn, mt)
+        handler.handle_polling = lambda fns, mt: 'Got update for %s since %s' % (','.join(fns), mt)
         with patch('os.path.isdir', lambda dir: dir == self.root):
             body = handler.do_GET_or_HEAD()
         expected_fn = self.filepath('a.txt')
@@ -156,11 +158,30 @@ class TestMyRequestHandler(unittest.TestCase):
         handler = MyRequestHandlerForTests()
         handler.path = '/polling?pathname=/&mtime=12345'
         handler.server.renderer.root = self.filepath('a.txt')
-        handler.handle_polling = lambda fn, mt: 'Got update for %s since %s' % (fn, mt)
+        handler.handle_polling = lambda fns, mt: 'Got update for %s since %s' % (','.join(fns), mt)
         body = handler.do_GET_or_HEAD()
         expected_fn = self.filepath('a.txt')
         self.assertEqual(body, 'Got update for %s since 12345' % expected_fn)
 
+    def test_do_GET_or_HEAD_polling_of_command_with_watch_files(self):
+        handler = MyRequestHandlerForTests()
+        handler.path = '/polling?pathname=/&mtime=12345'
+        handler.server.renderer.command = 'python setup.py --long-description'
+        handler.server.renderer.watch = ['setup.py', 'README.rst']
+        handler.handle_polling = lambda fns, mt: 'Got update for %s since %s' % (','.join(fns), mt)
+        body = handler.do_GET_or_HEAD()
+        self.assertEqual(body, 'Got update for setup.py,README.rst since 12345')
+
+    def test_do_GET_or_HEAD_polling_of_root_with_watch_files(self):
+        handler = MyRequestHandlerForTests()
+        handler.path = '/polling?pathname=/&mtime=12345'
+        handler.server.renderer.root = self.filepath('a.txt')
+        handler.server.renderer.watch = ['my.css']
+        handler.handle_polling = lambda fns, mt: 'Got update for %s since %s' % (','.join(fns), mt)
+        body = handler.do_GET_or_HEAD()
+        expected_fn = self.filepath('a.txt')
+        self.assertEqual(body, 'Got update for %s,my.css since 12345' % expected_fn)
+
     def test_do_GET_or_HEAD_prevent_sandbox_climbing_attacks(self):
         handler = MyRequestHandlerForTests()
         handler.path = '/polling?pathname=../../../etc/passwd&mtime=12345'
@@ -188,7 +209,7 @@ class TestMyRequestHandler(unittest.TestCase):
             handler = MyRequestHandlerForTests()
             handler.path = '/' + filename
             handler.server.renderer.root = self.filepath('file.txt')
-            handler.handle_rest_file = lambda fn: 'HTML for %s' % fn
+            handler.handle_rest_file = lambda fn, watch=None: 'HTML for %s' % fn
             body = handler.do_GET_or_HEAD()
             self.assertEqual(body, 'HTML for %s' % self.filepath(filename))
 
@@ -206,7 +227,7 @@ class TestMyRequestHandler(unittest.TestCase):
         with patch('time.sleep') as sleep:
             stat = {filename: [Mock(st_mtime=123455), Mock(st_mtime=123456)]}
             with patch('os.stat', lambda fn: stat[fn].pop(0)):
-                handler.handle_polling(filename, 123455)
+                handler.handle_polling([filename], 123455)
             sleep.assert_called_once_with(0.2)
         self.assertEqual(handler.status, 200)
         self.assertEqual(handler.headers['Cache-Control'],
@@ -219,7 +240,7 @@ class TestMyRequestHandler(unittest.TestCase):
         filename = os.path.join(os.path.dirname(__file__), '__init__.py')
         stat = {filename: [Mock(st_mtime=123456)]}
         with patch('os.stat', lambda fn: stat[fn].pop(0)):
-            handler.handle_polling(filename, 123455)
+            handler.handle_polling([filename], 123455)
         self.assertEqual(handler.log,
              ['connection reset by peer (client closed "/polling?pathname=__init__.py&mtime=123455" before acknowledgement)'])
 
@@ -231,7 +252,7 @@ class TestMyRequestHandler(unittest.TestCase):
                                self._raise_oserror,
                                lambda: Mock(st_mtime=123456)]}
             with patch('os.stat', lambda fn: stat[fn].pop(0)()):
-                handler.handle_polling(filename, 123455)
+                handler.handle_polling([filename], 123455)
         self.assertEqual(handler.status, 200)
         self.assertEqual(handler.headers['Cache-Control'],
                          "no-cache, no-store, max-age=0")
@@ -297,6 +318,21 @@ class TestMyRequestHandler(unittest.TestCase):
         self.assertTrue(body.startswith(b'HTML for'))
         self.assertTrue(body.endswith(('with AJAX poller for %s' % mtime).encode()))
 
+    def test_handle_rest_file_extra_watch(self):
+        handler = MyRequestHandlerForTests()
+        filename = os.path.join(os.path.dirname(__file__), '__init__.py')
+        mtime = os.stat(filename).st_mtime
+        with patch('os.stat', lambda fn: {filename: Mock(st_mtime=mtime),
+                                          'my.css': Mock(st_mtime=mtime + 1)}[fn]):
+            body = handler.handle_rest_file(filename, watch=['my.css'])
+        self.assertEqual(handler.status, 200)
+        self.assertTrue(body.endswith(('with AJAX poller for %s' % (mtime + 1)).encode()))
+        with patch('os.stat', lambda fn: {filename: Mock(st_mtime=mtime),
+                                          'my.css': Mock(st_mtime=mtime - 1)}[fn]):
+            body = handler.handle_rest_file(filename, watch=['my.css'])
+        self.assertEqual(handler.status, 200)
+        self.assertTrue(body.endswith(('with AJAX poller for %s' % mtime).encode()))
+
     def test_handle_rest_file_error(self):
         handler = MyRequestHandlerForTests()
         handler.path = '/nosuchfile.txt'
@@ -333,6 +369,22 @@ class TestMyRequestHandler(unittest.TestCase):
                          str(len(body)))
         self.assertEqual(handler.headers['Cache-Control'],
                          "no-cache, no-store, max-age=0")
+        self.assertFalse('X-Restview-Mtime' in handler.headers)
+        self.assertTrue(b'cat: README.rst: no such file' in body,
+                        body)
+
+    def test_handle_command_returns_error_with_watch_files(self):
+        handler = MyRequestHandlerForTests()
+        with patch('subprocess.Popen', PopenStub('', 'cat: README.rst: no such file', 1)):
+            body = handler.handle_command('cat README.rst', watch=['README.rst'])
+        self.assertEqual(handler.status, 200)
+        self.assertEqual(handler.headers['Content-Type'],
+                         "text/html; charset=UTF-8")
+        self.assertEqual(handler.headers['Content-Length'],
+                         str(len(body)))
+        self.assertEqual(handler.headers['Cache-Control'],
+                         "no-cache, no-store, max-age=0")
+        self.assertTrue('X-Restview-Mtime' in handler.headers)
         self.assertTrue(b'cat: README.rst: no such file' in body,
                         body)
 
@@ -457,6 +509,7 @@ def doctest_RestViewer_rest_to_html():
         ...
         ... This is an inline literal: ``README.txt``.
         ... ''', settings={'cloak_email_addresses': True}).strip())
+        ... # doctest: +ELLIPSIS,+REPORT_NDIFF
         <?xml version="1.0" encoding="utf-8" ?>
         <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
         <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
@@ -465,13 +518,10 @@ def doctest_RestViewer_rest_to_html():
         ...
         <title>example</title>
         <style type="text/css">
-        <BLANKLINE>
-        /*
-         * Stylesheet for ReStructuredText by Marius Gedminas.
-         * (I didn't like the default one)
         ...
-        </style>
-        <style type="text/css">
+        /*
+         * Stylesheet overrides for ReSTview
+         */
         ...
         </style>
         </head>
@@ -508,10 +558,11 @@ def doctest_RestViewer_rest_to_html_css_url():
     means it's hard to override them!
 
         >>> viewer = RestViewer('.')
-        >>> viewer.css_url = 'http://example.com/my.css'
+        >>> viewer.stylesheets = 'http://example.com/my.css'
         >>> print(viewer.rest_to_html(b'''
         ... Some text
         ... ''').strip())
+        ... # doctest: +ELLIPSIS,+REPORT_NDIFF
         <?xml version="1.0" encoding="utf-8" ?>
         <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
         <html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
@@ -543,23 +594,19 @@ def doctest_RestViewer_rest_to_html_strict_and_error_handling():
         >>> stderr = stderr_patcher.start()
 
         >>> viewer = RestViewer('.')
-        >>> viewer.css_path = viewer.css_url = None
+        >>> viewer.stylesheets = None
         >>> viewer.strict = True
         >>> print(viewer.rest_to_html(b'''
         ... Some text with an `error
-        ... ''').strip())
+        ... ''', mtime=1364808683).strip())
+        ... # doctest: +ELLIPSIS,+REPORT_NDIFF
         <!DOCTYPE html>
         <html>
         <head>
         <title>SystemMessage</title>
         <style type="text/css">
         pre.error {
-            border-left: 3px double red;
-            margin-left: 19px;
-            padding-left: 19px;
-            padding-top: 10px;
-            padding-bottom: 10px;
-            color: red;
+            ...
         }
         </style>
         </head>
@@ -573,6 +620,11 @@ def doctest_RestViewer_rest_to_html_strict_and_error_handling():
         Some text with an `error
         <BLANKLINE>
         </pre>
+        <BLANKLINE>
+        <script type="text/javascript">
+        var mtime = '1364808683';
+        ...
+        </script>
         </body>
         </html>
 
@@ -588,7 +640,7 @@ def doctest_RestViewer_rest_to_html_pypi_strict_and_error_handling():
         >>> stderr = stderr_patcher.start()
 
         >>> viewer = RestViewer('.')
-        >>> viewer.css_path = viewer.css_url = None
+        >>> viewer.stylesheets = None
         >>> viewer.pypi_strict = True
         >>> print(viewer.rest_to_html(b'''
         ... Hello
@@ -599,18 +651,14 @@ def doctest_RestViewer_rest_to_html_pypi_strict_and_error_handling():
         ... Here is a `relative link <example.com>`__.
         ...
         ... ''').strip())
+        ... # doctest: +ELLIPSIS,+REPORT_NDIFF
         <!DOCTYPE html>
         <html>
         <head>
         <title>TransformError</title>
         <style type="text/css">
         pre.error {
-            border-left: 3px double red;
-            margin-left: 19px;
-            padding-left: 19px;
-            padding-top: 10px;
-            padding-bottom: 10px;
-            color: red;
+            ...
         }
         </style>
         </head>
@@ -675,6 +723,7 @@ def doctest_RestViewer_inject_ajax_adds_ajax():
         ... </body>
         ... </html>
         ... ''', mtime=1364808683).strip())
+        ... # doctest: +ELLIPSIS,+REPORT_NDIFF
         <html>
         <head>
         <title>Title</title>
@@ -683,26 +732,8 @@ def doctest_RestViewer_inject_ajax_adds_ajax():
         <p>Some body text</p>
         <BLANKLINE>
         <script type="text/javascript">
-        var xmlHttp = null;
-        window.onload = function () {
-            setTimeout(function () {
-                if (window.XMLHttpRequest) {
-                    xmlHttp = new XMLHttpRequest();
-                } else if (window.ActiveXObject) {
-                    xmlHttp = new ActiveXObject('Microsoft.XMLHTTP');
-                }
-                xmlHttp.onreadystatechange = function () {
-                    if (xmlHttp.readyState == 4 && xmlHttp.status == '200') {
-                        window.location.reload(true);
-                    }
-                }
-                xmlHttp.open('HEAD', '/polling?pathname=' + location.pathname + '&mtime=1364808683', true);
-                xmlHttp.send(null);
-            }, 0);
-        }
-        window.onbeforeunload = function () {
-            xmlHttp.abort();
-        }
+        var mtime = '1364808683';
+        ...
         </script>
         </body>
         </html>
